# -*- coding: utf-8 -*-
"""
Created on Fri Jan 10 14:00:59 2025

@author: mstep
"""

# Himeji image:
# https://commons.wikimedia.org/wiki/File:Himeji_Castle_0804_1.jpg

# Click sound:
# https://freesound.org/people/kwahmah_02/sounds/260609/

# Tada sound:
# https://freesound.org/people/plasterbrain/sounds/397354/

#%%
'''The custom SettingsDialog class.'''
import tkinter as tk
from tkinter import messagebox

class SettingsDialog(tk.Toplevel):
    def __init__(self, master, piece_size=50, min_size=20):
        super().__init__(master)
        self.transient(master)  # Keep dialog above master.
        self.minsize(300, 200)
        self.title('Settings')
        self.min_size = min_size
        self.piece_size = None

        # Make a frame to hold the dialog's main body.
        main_frame = tk.Frame(self)
        main_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=10, pady=10)

        # Build the dialog's widgets.
        piece_size_label = tk.Label(main_frame, text='Piece size (pixels):')
        piece_size_label.grid(row=0, column=0, sticky=tk.W, padx=5, pady=5)
        self.piece_size_entry = tk.Entry(main_frame)
        self.piece_size_entry.insert(0, f'{piece_size}')
        self.piece_size_entry.selection_range(0, tk.END)
        self.piece_size_entry.grid(row=0, column=1, sticky=tk.W, padx=5, pady=5)

        # Make OK and Cancel buttons.
        button_frame = tk.Frame(self, height=5)
        button_frame.pack(side=tk.RIGHT, fill=tk.X, expand=False, padx=10, pady=10)

        ok_button = tk.Button(button_frame, text='OK', width=7,
            command=self.ok_click)
        ok_button.pack(side=tk.LEFT, padx=10, pady=10)
        self.bind('<Return>', lambda event: self.ok_click())

        cancel_button = tk.Button(button_frame, text='Cancel', width=7,
            command=self.cancel_click)
        cancel_button.pack(side=tk.LEFT, padx=10, pady=10)
        self.bind('<Escape>', lambda event: self.cancel_click())

        self.piece_size_entry.focus_set()

        # Grab events for this dialog.
        self.grab_set()

    def cancel_click(self):
        '''Close the dialog.'''
        self.destroy()

    def ok_click(self):
        '''Validate the user's entries and close the dialog.'''
        if not self.validate(): return

        # Close the dialog.
        self.withdraw()
        self.update_idletasks()
        self.cancel_click()

    def validate(self):
        '''
        Validate the user's entries.
        Display an error message if appropriate.
        Return True if the entries are valid, False otherwise.
        '''
        try:
            piece_size = int(self.piece_size_entry.get())
        except:
            piece_size = 0
        if piece_size < self.min_size:
            messagebox.showerror('Error',
                f'The size should be at least {self.min_size} pixels.')
            self.piece_size_entry.focus_set()
            return False

        # Save the validated values.
        self.piece_size = piece_size
        return True

    def show(self):
        '''Display the dialog, wait for it to end, and return its results.'''
        self.wait_window()
        return self.piece_size

#%%

def divide_evenly(number, size):
    '''
    Divide number into parts that are roughly size in size.
    Example: Divide 310 into parts of size roughly 50: [52, 52, 52, 52, 51, 51]
    '''
    if size > number:
        size = number
    num_parts = number // size       # See how many parts there will be.
    base_size = number // num_parts  # Get "normal" part size.
    remainder = number % num_parts   # Get leftover after making the parts.

    # Create the parts.
    parts = [base_size] * num_parts

    # Make some parts bigger to use up the remainder.
    for i in range(remainder):
        parts[i] += 1

    return parts

#%%
import math
import winsound # Windows only.

class PuzzlePiece:
    def __init__(self, puzzle_image, x, y, width, height):
        '''Save this piece's part of the puzzle.'''
        self.x0 = x    # The piece's home location.
        self.y0 = y
        self.x = x     # The piece's current location.
        self.y = y
        self.width = width
        self.height = height
        self.image = puzzle_image.crop((x, y, x + width, y + height))
        self.is_home = False

    def __repr__(self):
        '''Return a string representation for the piece.'''
        return(f'PuzzlePiece({self.x}, {self.y}, {self.width}, {self.height})')

    def draw(self, dr, table_image):
        '''Draw this piece at its current position on the table image.'''
        table_image.paste(self.image, (self.x, self.y))

        # If the piece is not home, outline it.
        if not self.is_home:
            THICKNESS = 3
            x1 = self.x + self.width
            y1 = self.y + self.height
            dr.rectangle(((self.x, self.y), (x1, y1)),
                outline='yellow', width=THICKNESS)

    def is_at(self, x, y):
        '''Return True if the piece is at this position.'''
        # Once we're locked, it cannot move.
        if self.is_home: return False

        if x < self.x: return False
        if y < self.y: return False
        if x > self.x + self.width - 1: return False
        if y > self.y + self.height - 1: return False
        return True

    def drop(self):
        '''Drop this piece at its current position.'''
        CLOSE_ENOUGH = 10 # Distance in pixels.

        # See if x and y are within CLOSE of our home location.
        if math.isclose(self.x0, self.x, abs_tol=CLOSE_ENOUGH) and \
           math.isclose(self.y0, self.y, abs_tol=CLOSE_ENOUGH):
            # Go home.
            self.x, self.y = self.x0, self.y0

            # Lock the piece.
            self.is_home = True

            # Return True to indcate that the piece is locked.
            return True

#%%

import tkinter as tk
from tkinter import filedialog as fd
from PIL import Image, ImageTk, ImageDraw
import random

class JigsawApp:
    def __init__(self):
        self.window = tk.Tk()

        self.window.title('jigsaw')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('300x300')
        self.window.resizable(False, False)

        # Set initial data.
        self.puzzle_image = None
        self.piece_size = 100
        self.selected_piece = None

        # Build the menu.
        self.build_menus()

        # Build other stuff...
        self.canvas = tk.Canvas(self.window)
        self.canvas.pack(side=tk.TOP, padx=5, pady=5, fill=tk.BOTH, expand=True)

        # Track Button-1 down, Button-1 motion, and Button-1 up.
        self.canvas.bind("<Button-1>", self.mouse_down)
        self.canvas.bind("<B1-Motion>", self.mouse_move)
        self.canvas.bind("<ButtonRelease-1>", self.mouse_up)

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def build_menus(self):
        '''Build the menu bar.'''
        menu_bar = tk.Menu(self.window)
        self.window.config(menu=menu_bar)
        self.build_file_menu(menu_bar)
        self.build_config_menu(menu_bar)

    def build_file_menu(self, menu_bar):
        '''Build the File menu.'''
        file_menu = tk.Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label='File', menu=file_menu)

        # Open
        self.open_photo_image = ImageTk.PhotoImage(file='open.png')
        file_menu.add_command(label='Open Image...', accelerator='Ctrl+O',
            compound='left',
            command=self.mnu_open_image, image=self.open_photo_image)
        self.window.bind('<Control-o>', lambda event: self.mnu_open_image())

        # Note: If you don't include an image, the menu item doesn't
        # quite line up with items that do have an image. You can fix
        # it by including a blank image.

        file_menu.add_separator()

        # Exit
        self.exit_photo_image = ImageTk.PhotoImage(file='exit.png')
        file_menu.add_command(label='Exit', compound='left',
            command=self.mnu_exit, image=self.exit_photo_image)

    def build_config_menu(self, menu_bar):
        '''Build the Configuration menu.'''
        config_menu = tk.Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label='Configuration', menu=config_menu)

        # Settings
        self.settings_image = ImageTk.PhotoImage(file='settings.png')
        config_menu.add_command(label='Settings...', compound='left',
            command=self.mnu_settings, image=self.settings_image)

    def mnu_open_image(self):
        self.puzzle_image = self.get_image_file('Load Image')
        self.disassemble_image()
        self.show_image()

    def mnu_exit(self):
        self.kill_callback()

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def mnu_settings(self):
        '''Let the user set the piece size.'''
        size = SettingsDialog(self.window, self.piece_size, min_size=20).show()
        if size is not None:
            self.piece_size = size
            self.disassemble_image()
            self.show_image()

    def get_image_file(self, title):
        '''
        Let the user pick an image file.
        Then load and return it in an RGB PIL Image.
        '''
        filetypes = (
            ('Image Files', '*.png;*.jpg;*.bmp;*.gif;*.tif'),
            ('All files', '*.*')    
        )
        filename = fd.askopenfilename(title=title, filetypes=filetypes)
        if not filename: return None

        try:
            return Image.open(filename)
        except Exception as e:
            tk.messagebox.showinfo('Open Error',
                f'Error loading image file.\n{e}')
            return None

    def disassemble_image(self):
        '''Break the image into puzzle pieces and return the pieces.'''
        if self.puzzle_image is None: return

        # Get the kitchen table's image size allowing extra space
        # on the sides for unplaced pieces.
        WIDTH_FACTOR = 2
        HEIGHT_ADDITION = 100
        self.table_width = round(self.puzzle_image.width * WIDTH_FACTOR)
        self.table_height = self.puzzle_image.height + HEIGHT_ADDITION

        # Calculate row and column sizes.
        row_heights = divide_evenly(self.puzzle_image.height, self.piece_size)
        col_widths = divide_evenly(self.puzzle_image.width, self.piece_size)

        # Make the pieces.
        self.pieces = []
        y = 0
        for height in row_heights:
            x = 0
            for width in col_widths:
                # Make the piece at (x, y).
                self.pieces.append(PuzzlePiece(self.puzzle_image,
                    x, y, width, height))
                x += width
            y += height

        # Randomly position the pieces.
        self.randomize_pieces()

    def randomize_pieces(self):
        '''Randomly position the pieces.'''
        image_x0 = (self.table_width - self.puzzle_image.width) // 2
        image_x1 = image_x0 + self.puzzle_image.width
        image_y0 = (self.table_height - self.puzzle_image.height) // 2
        MARGIN = 5
        xmin = MARGIN
        ymin = MARGIN
        xmax = image_x0 - self.pieces[0].width - MARGIN
        if xmax < xmin + 50:    # Can happen for very large sizes.
            xmax = xmin + 50
        ymax = self.table_height - self.pieces[0].height - MARGIN
        for piece in self.pieces:
            # Move this piece's home position so it's on the table.
            piece.x0 += image_x0
            piece.y0 += image_y0

            # Position this piece.
            x = random.randint(xmin, xmax)
            # See if it's to the right of the puzzle image.
            if random.randint(0, 1) == 0:
                # Move to the right.
                x += image_x1
            y = random.randint(ymin, ymax)
            piece.x = x
            piece.y = y
        self.num_unplaced_pieces = len(self.pieces)

    def show_image(self):
        '''Build and display the kitchen table image.'''
        if self.puzzle_image == None: return

        # Make the kitchen table image.
        table_image = Image.new('RGB',
            (self.table_width, self.table_height), 'lightgray')

        # Make an ImageDraw object to draw on the table image.
        dr = ImageDraw.Draw(table_image)

        # Make the pieces draw themselves.
        for piece in self.pieces:
            piece.draw(dr, table_image)

        # Outline the puzzle area.
        x0 = (self.table_width - self.puzzle_image.width) // 2
        y0 = (self.table_height- self.puzzle_image.height) // 2
        x1 = x0 + self.puzzle_image.width - 1
        y1 = y0 + self.puzzle_image.height - 1
        dr.rectangle(((x0, y0), (x1, y1)), outline='red', width=5)

        # Convert into a PhotoImage.
        self.photo_image = ImageTk.PhotoImage(table_image)

        # Display it.
        self.canvas.delete(tk.ALL)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.photo_image)
        self.canvas.config(width=self.table_width, height=self.table_height)

        # Size the program to fit.
        self.window.geometry(f'{self.table_width}x{self.table_height}')

    def piece_at(self, x, y):
        '''Return the puzzle piece (if any) at this position.'''
        # Check in reverse order so pieces drawn last (so they are drawn
        # on top) are checked first.
        for piece in reversed(self.pieces):
            if piece.is_at(x, y):
                return piece
        return None # We didn't find a piece here.

    def mouse_down(self, event):
        '''If a puzzle piece is below the mouse, start moving it.'''
        # Do nothing if we haven't loaded a puzzle yet.
        if self.pieces is None: return

        # See if there is a piece at this position.
        self.selected_piece = self.piece_at(event.x, event.y)

        # See if we found a piece here.
        if self.selected_piece is not None:
            # Move it to the end of the list so it is drawn on top.
            self.pieces.remove(self.selected_piece)
            self.pieces.append(self.selected_piece)

            # Redraw.
            self.show_image()

            # Save the mouse's position with respect to the root window.
            self.start_x = event.x_root
            self.start_y = event.y_root

            # Save the piece's current location.
            self.piece_x = self.selected_piece.x
            self.piece_y = self.selected_piece.y

            # Calculate minimum X and Y coordinates.
            self.max_x = self.canvas.winfo_width() - self.selected_piece.width
            self.max_y = self.canvas.winfo_height() - self.selected_piece.height

    def mouse_move(self, event):
        '''Move the selected puzzle piece.'''
        # Ignore extraneous events.
        if self.selected_piece is None: return

        # Move the piece.
        x = self.piece_x + (event.x_root - self.start_x)
        y = self.piece_y + (event.y_root - self.start_y)
        if x > self.max_x: x = self.max_x
        if y > self.max_y: y = self.max_y
        if x < 0: x = 0
        if y < 0: y = 0
        self.selected_piece.x = x
        self.selected_piece.y = y

        # Redraw.
        self.show_image()

    def mouse_up(self, event):
        '''Stop moving the selected puzzle piece.'''
        # Ignore extraneous events.
        if self.selected_piece is None: return

        # Drop the piece at its current position.
        if self.selected_piece.drop():
            # This piece was locked.
            # Move it to the bottom of the stacking order.
            self.pieces.remove(self.selected_piece)
            self.pieces.insert(0, self.selected_piece)

            # See if we're done.
            self.num_unplaced_pieces -= 1
            if self.num_unplaced_pieces == 0:
                # Play the "game over" sound.
                winsound.PlaySound('tada.wav',
                    winsound.SND_FILENAME or winsound.SND_ASYNC)
            else:
                # Play the "locked" sound.
                winsound.PlaySound('click.wav',
                    winsound.SND_FILENAME or winsound.SND_ASYNC)

        # We are no longer dragging the piece.        
        self.selected_piece = None

        # Redraw.
        self.show_image()

# In[19]:
JigsawApp()
print('Done')
